from __future__ import annotations

import configparser
import logging
from typing import TYPE_CHECKING, Iterable

from configparser_override._strategy_factory import StrategyFactory
from configparser_override.convert import ConfigConverter

if TYPE_CHECKING:
    from _typeshed import StrOrBytesPath

    from configparser_override._override_strategy import Strategy
    from configparser_override.types import _dataclass, _optionxform_fn

logger = logging.getLogger(__name__)


class ConfigParserOverride:
    def __init__(
        self,
        env_prefix: str = "",
        create_new_from_env_prefix: bool = True,
        create_new_from_direct: bool = True,
        config_parser: configparser.ConfigParser | None = None,
        case_sensitive_overrides: bool = False,
        optionxform: _optionxform_fn | None = None,
        **overrides: str,
    ):
        """
        Initialize the ConfigParserOverride.

        :param env_prefix: Optional prefix for environment variables,
            defaults to an empty string.
        :type env_prefix: str, optional
        :param create_new_from_env_prefix: Flag to create new configuration
            options from environment variables.
        :type create_new_from_env_prefix: bool, optional
        :param create_new_from_direct: Flag to create new configuration
            options from direct overrides.
        :type create_new_from_direct: bool, optional
        :param config_parser: Optional ConfigParser object to be used,
            defaults to None.
        :type config_parser: configparser.ConfigParser, optional
        :param case_sensitive_overrides: Flag to indicate if overrides should
            be case sensitive.
        :type case_sensitive_overrides: bool, optional
        :param optionxform: Optional function to transform option strings.
        :type optionxform: _optionxform_fn | None, optional
        :param overrides: Keyword arguments to directly override configuration values.
        :type overrides: dict[str, str | None]
        """

        self.env_prefix = env_prefix
        self.create_new_from_env_prefix = create_new_from_env_prefix
        self.create_new_from_direct = create_new_from_direct
        self.case_sensitive_overrides = case_sensitive_overrides
        self.optionxform = optionxform
        self.overrides = overrides

        # Configure ConfigParser and align optionxform for consistency in later
        # inferance for overrides
        if config_parser is None:
            self._config = configparser.ConfigParser()
            if self.optionxform is not None:
                self._config.optionxform = self.optionxform  # type: ignore
        else:
            self._config = config_parser
            self.optionxform = self._config.optionxform

    def _get_override_strategy(self) -> Strategy:
        """
        Get the appropriate override strategy based on initialization parameters.

        :return: The appropriate strategy instance.
        :rtype: Strategy
        """
        return StrategyFactory(
            self._config,
            self.env_prefix,
            self.create_new_from_env_prefix,
            self.create_new_from_direct,
            self.overrides,
            self.case_sensitive_overrides,
            self.optionxform,
        ).get_strategy()

    def read(
        self,
        filenames: StrOrBytesPath | Iterable[StrOrBytesPath],
        encoding: str | None = None,
    ) -> configparser.ConfigParser:
        """
        Read configuration from one or more files and override with environment
        variables if set.

        This method is a wrapper around :py:meth:`configparser.ConfigParser.read` that
        reads the specified filenames in order. After reading the files, it overrides
        configuration values with corresponding environment variables and direct
        overrides passed during initialization.

        :param filenames: A single filename or an iterable of filenames to read.
        :type filenames: :py:class:`_typeshed.StrOrBytesPath` or
            Iterable[:py:class:`_typeshed.StrOrBytesPath`]
        :param encoding: The encoding to use for reading the files, defaults to None.
        :type encoding: str, optional
        :return: The :py:class:`configparser.ConfigParser` object with the loaded and
            possibly overridden configuration.
        :rtype: :py:class:`configparser.ConfigParser`

        **Examples:**

        .. code-block:: python

            >>> parser_override = ConfigParserOverride(test_option='value')
            >>> config = parser_override.read(['example.ini'])
            >>> config.get('DEFAULT', 'test_option')
            'value'


        """
        self._config.read(filenames=filenames, encoding=encoding)
        strategy = self._get_override_strategy()
        strategy.execute()
        return self.config

    @property
    def config(self) -> configparser.ConfigParser:
        """
        Property to access the configuration.

        This can be used to modify the property of the configparser object and
        also set and get options manually.

        :return: The :py:class:`configparser.ConfigParser` object
            with the configuration.
        :rtype: :py:class:`configparser.ConfigParser`

        **Examples:**

        Get an option after parsing and overrides:

        .. code-block:: python

            >>> config = ConfigParserOverride(test_option='value')
            >>> config.read(['example.ini'])
            >>> config.get('DEFAULT', 'test_option')
            'value'

        Can also be used like just like regular ConfigParser:

        .. code-block:: python

            >>> parser_override = ConfigParserOverride()
            >>> config = parser_override.config
            >>> config.set('section', 'option', 'value')
            >>> config.get('section', 'option')
            'value'

        """
        return self._config

    def to_dataclass(self, dataclass: _dataclass) -> _dataclass:
        """
        Convert the configuration data to a dataclass instance.

        Can be useful to validate that the configuration adheres to the expected format
        and leverage various typing frameworks, eg. integrations with your text editor.

        :param dataclass: The dataclass type to convert the configuration data into.
        :type dataclass: _dataclass
        :return: An instance of the dataclass populated with the configuration data.
        :rtype: _dataclass

        **Examples:**

        .. code-block:: python

            >>> from dataclasses import dataclass

            >>> @dataclass
            ... class Section1:
            ...     key: str

            >>> @dataclass
            ... class ExampleConfig:
            ...     section1: Section1

            >>> config_parser_override = ConfigParserOverride(section1__key="a string")
            >>> config_parser_override.read()
            >>> config_as_dataclass = config_parser_override.to_dataclass(ExampleConfig)
            >>> assert config_as_dataclass.section1.key == "a string" # True
        """

        return ConfigConverter(self._config).config_to_dataclass(dataclass)



import configparser
import platform

import pytest

from configparser_override._override_strategy import (
    NewOptionsFromDirectAndEnvStrategy,
    NewOptionsFromDirectStrategy,
    NewOptionsFromEnvStrategy,
    NoNewOptionsStrategy,
    _lowercase_optionxform,
)
from configparser_override.exceptions import SectionNotFound
from tests._constants import TEST_ENV_PREFIX


def test_lowercase_optionxform():
    assert _lowercase_optionxform("TEST") == "test"
    assert _lowercase_optionxform("TesT") == "test"
    assert _lowercase_optionxform("test") == "test"


def test_no_prefix_no_new_strategy_executes_overrides():
    config = configparser.ConfigParser()
    config.add_section("SECTION1")
    config.set("SECTION1", "option1", "value1")

    overrides = {"SECTION1__option1": "new_value1"}
    strategy = NoNewOptionsStrategy(config, "", overrides)
    strategy.execute()

    assert config.get("SECTION1", "option1") == "new_value1"


def test_no_prefix_new_direct_strategy_creates_new_options():
    config = configparser.ConfigParser()
    config.add_section("SECTION1")
    config.set("SECTION1", "option1", "value1")

    overrides = {"section2__option2": "new_value2"}
    strategy = NewOptionsFromDirectStrategy(config, "", overrides)
    strategy.execute()

    assert config.get("section2", "option2") == "new_value2"


def test_no_prefix_new_direct_strategy_creates_new_options_with_sensetive_case():
    config = configparser.ConfigParser()
    config.add_section("SECTION1")
    config.set("SECTION1", "option1", "value1")

    overrides = {"SECTION2__OPTION2": "new_value2"}
    strategy = NewOptionsFromDirectStrategy(config, "", overrides, True)
    strategy.execute()

    assert config.get("SECTION2", "OPTION2") == "new_value2"


def test_prefix_new_direct_overrides_existing_options(monkeypatch):
    config = configparser.ConfigParser()
    config.add_section("SECTION1")
    config.set("SECTION1", "option1", "value1")
    config.set("SECTION1", "option2", "value2")

    monkeypatch.setenv(f"{TEST_ENV_PREFIX}SECTION1__OPTION1", "env_value1")
    overrides = {"section1__option3": "direct_value3"}
    strategy = NewOptionsFromDirectStrategy(config, TEST_ENV_PREFIX, overrides)
    strategy.execute()

    assert config.get("SECTION1", "option1") == "env_value1"
    assert config.get("SECTION1", "option2") == "value2"
    assert config.get("SECTION1", "option3") == "direct_value3"


def test_prefix_no_new_strategy_overrides_existing_options(monkeypatch):
    config = configparser.ConfigParser()
    config.add_section("SECTION1")
    config.set("SECTION1", "option1", "value1")

    monkeypatch.setenv(f"{TEST_ENV_PREFIX}SECTION1__OPTION1", "env_value1")
    overrides = {}
    strategy = NoNewOptionsStrategy(config, TEST_ENV_PREFIX, overrides)
    strategy.execute()

    assert config.get("SECTION1", "option1") == "env_value1"


def test_prefix_new_env_strategy_creates_new_options_from_env(monkeypatch):
    config = configparser.ConfigParser()

    monkeypatch.setenv(f"{TEST_ENV_PREFIX}SECTION1__OPTION1", "env_value1")
    overrides = {}
    strategy = NewOptionsFromEnvStrategy(config, TEST_ENV_PREFIX, overrides)
    strategy.execute()

    assert config.get("section1", "option1") == "env_value1"


def test_prefix_new_env_strategy_creates_new_options_from_env_case_sensetive_upper(
    monkeypatch,
):
    config = configparser.ConfigParser()

    monkeypatch.setenv(f"{TEST_ENV_PREFIX}SECTION1__OPTION1", "env_value1")
    overrides = {}
    strategy = NewOptionsFromEnvStrategy(config, TEST_ENV_PREFIX, overrides, True)
    strategy.execute()

    assert config.get("SECTION1", "option1") == "env_value1"


def test_prefix_new_env_strategy_creates_new_options_from_env_case_sensetive_lower(
    monkeypatch,
):
    config = configparser.ConfigParser()

    monkeypatch.setenv(f"{TEST_ENV_PREFIX}section1__option1", "env_value1")
    overrides = {}
    strategy = NewOptionsFromEnvStrategy(config, TEST_ENV_PREFIX, overrides, True)
    strategy.execute()

    p = platform.system()
    if p == "Windows":
        assert (
            config.get("SECTION1", "option1") == "env_value1"
        )  # Env var stored as capital for win
    elif p == "Linux" or p == "Darwin":
        assert config.get("section1", "option1") == "env_value1"


def test_prefix_new_env_new_direct_strategy_creates_both(monkeypatch):
    config = configparser.ConfigParser()

    monkeypatch.setenv(f"{TEST_ENV_PREFIX}SECTION1__OPTION1", "env_value1")
    overrides = {"SECTION2__option2": "new_value2"}
    strategy = NewOptionsFromDirectAndEnvStrategy(config, TEST_ENV_PREFIX, overrides)
    strategy.execute()

    assert config.get("section1", "option1") == "env_value1"
    assert config.get("section2", "option2") == "new_value2"


def test_parse_key_case_insensitive():
    config = configparser.ConfigParser()
    strategy = NoNewOptionsStrategy(config, "", {})
    section, option = strategy.parse_key("SECTION__OPTION")
    assert section == "SECTION"
    assert option == "option"


def test_decide_env_var_case_insensitive():
    config = configparser.ConfigParser()
    strategy = NoNewOptionsStrategy(config, "", {})
    env_var = strategy.decide_env_var("", "SECTION", "OPTION")
    assert env_var == "SECTION__OPTION"


def test_decide_env_var_case_sensitive():
    config = configparser.ConfigParser()
    strategy = NoNewOptionsStrategy(config, "", {}, case_sensitive_overrides=True)
    env_var = strategy.decide_env_var("", "section", "OPTION")
    assert env_var == "section__OPTION"


def test_decide_env_var_case_insensitive_prefix():
    config = configparser.ConfigParser()
    strategy = NoNewOptionsStrategy(config, "PREFIX_", {})
    env_var = strategy.decide_env_var("PREFIX_", "SECTION", "OPTION")
    assert env_var == "PREFIX_SECTION__OPTION"


def test_decide_env_var_case_sensitive_prefix():
    config = configparser.ConfigParser()
    strategy = NoNewOptionsStrategy(
        config, "PREFIX_", {}, case_sensitive_overrides=True
    )
    env_var = strategy.decide_env_var("PREFIX_", "section", "OPTION")
    assert env_var == "PREFIX_section__OPTION"


def test_has_section_case_insensitive():
    config = configparser.ConfigParser()
    config.add_section("section")
    strategy = NoNewOptionsStrategy(config, "", {})
    assert strategy.has_section("section")
    assert strategy.has_section("SECTION")


def test_has_section_case_sensitive():
    config = configparser.ConfigParser()
    config.add_section("SECTION")
    strategy = NoNewOptionsStrategy(config, "", {}, case_sensitive_overrides=True)
    assert strategy.has_section("SECTION")
    assert not strategy.has_section("section")


def test_get_existing_section_case_insensitive():
    config = configparser.ConfigParser()
    config.add_section("section")
    strategy = NoNewOptionsStrategy(config, "", {})
    assert strategy.get_existing_section_case_insensitive("SECTION") == "section"


def test_get_existing_section_case_sensitive():
    config = configparser.ConfigParser()
    config.add_section("SECTION")
    strategy = NoNewOptionsStrategy(config, "", {})
    with pytest.raises(SectionNotFound):
        strategy.get_existing_section_case_insensitive("NOT_A_SECTION")


def test_collect_env_vars_with_prefix(monkeypatch):
    monkeypatch.setenv("PREFIX_SECTION__OPTION", "value")
    strategy = NoNewOptionsStrategy(configparser.ConfigParser(), "PREFIX_", {})
    env_vars = strategy.collect_env_vars_with_prefix("PREFIX_")
    assert env_vars == {"SECTION__OPTION": "value"}



from __future__ import annotations

import enum
import logging
import os
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Mapping

from configparser_override.exceptions import SectionNotFound

if TYPE_CHECKING:
    import configparser

    from configparser_override.types import _optionxform_fn

logger = logging.getLogger(__name__)


def _lowercase_optionxform(optionstr: str) -> str:
    """
    Convert the given option string to lowercase.

    :param optionstr: The option string to be converted.
    :type optionstr: str
    :return: The lowercase version of the option string.
    :rtype: str
    """
    return optionstr.lower()


class Strategy(ABC):
    def __init__(
        self,
        config: configparser.ConfigParser,
        env_prefix: str,
        overrides: Mapping[str, str],
        case_sensitive_overrides: bool = False,
        optionxform_fn: _optionxform_fn | None = None,
    ):
        """
        Initialize the base Strategy class.

        :param config: The ConfigParser object to be used.
        :type config: configparser.ConfigParser
        :param env_prefix: Prefix for environment variables.
        :type env_prefix: str
        :param overrides: Mapping of override keys and values.
        :type overrides: Mapping[str, str | None]
        :param case_sensitive_overrides: Flag to indicate if overrides should
            be case sensitive.
        :type case_sensitive_overrides: bool, optional
        :param optionxform_fn: Optional function to transform option strings.
        :type optionxform_fn: _optionxform_fn | None, optional
        """
        self._config = config
        self._env_prefix = env_prefix
        self._overrides = overrides
        self.case_sensitive_overrides = case_sensitive_overrides
        if optionxform_fn is None:
            self.optionxform_fn = _lowercase_optionxform
        else:
            self.optionxform_fn = optionxform_fn

    @abstractmethod
    def execute(self):
        """Execute the strategy. Must be implemented by subclasses."""
        pass

    def collect_env_vars_with_prefix(self, prefix: str) -> dict[str, str]:
        """
        Collect environment variables that start with the given prefix.

        :param prefix: The prefix to filter environment variables.
        :type prefix: str
        :return: Dictionary of environment variables with the prefix removed.
        :rtype: dict[str, str]
        """
        if self.case_sensitive_overrides:
            return {
                key[len(prefix) :]: value
                for key, value in os.environ.items()
                if key.startswith(prefix)
            }
        return {
            key[len(prefix) :]: value
            for key, value in os.environ.items()
            if key.startswith(prefix.upper())
        }

    def decide_env_var(self, prefix: str, section: str, option: str) -> str:
        """
        Determine the appropriate environment variable name based on the given
        prefix, section, and option.

        :param prefix: The prefix for environment variables.
        :type prefix: str
        :param section: The section in the configuration.
        :type section: str
        :param option: The option in the configuration.
        :type option: str
        :return: The environment variable name.
        :rtype: str

        .. note::
            This method is aware of case-sensitivity setting

        """
        if self.case_sensitive_overrides:
            env_var = (
                f"{prefix}{section}__{option}"
                if section != self._config.default_section
                else f"{prefix}{option}"
            )
        else:
            env_var = (
                f"{prefix.upper()}{section.upper()}__{option.upper()}"
                if section.lower() != self._config.default_section.lower()
                else f"{prefix.upper()}{option.upper()}"
            )
        return env_var

    def parse_key(self, key: str) -> tuple[str, str]:
        """
        Parse a given key to extract the section and option.

        ConfigParser stores all options as lowercase by default, hence the option part
        is standardized to be lowercase unless a `optionxform` functions is specified.

        :param key: The key to parse.
        :type key: str
        :return: A tuple containing the section and option.
        :rtype: tuple[str, str]
        """
        parts = key.split("__", 1)
        if len(parts) == 1:
            return self._config.default_section, self.optionxform_fn(parts[0])
        return parts[0], self.optionxform_fn(parts[1])

    def has_section(self, section: str) -> bool:
        """
        Check if the section exists or is the default section.

        :param section: The section name to check.
        :type section: str
        :return: True if the section exists, False otherwise.
        :rtype: bool

        .. note::
            This method is aware of case-sensitivity setting

        """
        if self.case_sensitive_overrides:
            return (
                self._config.has_section(section)
                or section == self._config.default_section
            )
        return (
            section.lower() in (sect.lower() for sect in self._config.sections())
            or section.lower() == self._config.default_section.lower()
        )

    def get_existing_section_case_insensitive(self, section: str) -> str:
        """
        Get the existing section name in a case-insensitive manner.

        :param section: The section name to search for.
        :type section: str
        :return: The actual section name in the configuration.
        :rtype: str
        :raises SectionNotFound: If section is not found.
        """
        if section.lower() == self._config.default_section.lower():
            return self._config.default_section
        for sect in self._config.sections():
            if sect.lower() == section.lower():
                return sect
        raise SectionNotFound(f"Section {section} not found.")

    def override_and_add_new(self, key: str, value: str):
        section, option = self.parse_key(key)
        if self.case_sensitive_overrides:
            if not self.has_section(section):
                self._config.add_section(section=section)
            self._config.set(section=section, option=option, value=value)
        else:
            if not self.has_section(section):
                self._config.add_section(section=section.lower())
                self._config.set(section=section.lower(), option=option, value=value)
            else:
                _section = self.get_existing_section_case_insensitive(section)
                self._config.set(section=_section, option=option, value=value)

    def override_env(self, create_new_options: bool):
        """
        Override configuration values using environment variables.

        :param create_new_options: Flag to indicate if new options can be created.
        :type create_new_options: bool
        """
        if create_new_options:
            env_vars = (
                self.collect_env_vars_with_prefix(self._env_prefix)
                if self._env_prefix != ""
                else {}
            )
            for key, value in env_vars.items():
                self.override_and_add_new(key=key, value=value)
        else:
            for section in self._config.sections():
                for option in self._config[section]:
                    env_var = self.decide_env_var(self._env_prefix, section, option)
                    if env_var in os.environ:
                        _value = os.environ[env_var]
                        logger.debug(f"Override {section=}, {option=} with {env_var}")
                        self._config.set(section=section, option=option, value=_value)
                    else:
                        logger.debug(f"Environment variable {env_var} not set")

            _default_section = self._config.default_section
            for option in self._config.defaults():
                env_var = self.decide_env_var(
                    self._env_prefix, _default_section, option
                )
                if env_var in os.environ:
                    _value = os.environ[env_var]
                    logger.debug(
                        f"Override section={_default_section}, {option=} with {env_var}"
                    )
                    self._config.set(
                        section=_default_section, option=option, value=_value
                    )
                else:
                    logger.debug(f"Environment variable {env_var} not set")

    def override_direct(self, create_new_options: bool):
        """
        Override configuration values using direct overrides.

        :param create_new_options: Flag to indicate if new options can be created.
        :type create_new_options: bool
        """
        if create_new_options:
            for key, value in self._overrides.items():
                self.override_and_add_new(key=key, value=value)

        else:
            for key, value in self._overrides.items():
                section, option = self.parse_key(key)
                if self.case_sensitive_overrides:
                    if self.has_section(section) and self._config.has_option(
                        section, option
                    ):
                        logger.debug(
                            f"Override {section=}, {option=} with direct assignment"
                        )
                        self._config.set(section=section, option=option, value=value)
                    else:
                        logger.debug(
                            f"New direct assignment {section=} {option=} ignored"
                        )
                else:
                    if self.has_section(section):
                        section = self.get_existing_section_case_insensitive(section)
                        if self._config.has_option(section, option):
                            logger.debug(
                                f"Override {section=}, {option=} with direct assignment"
                            )
                            self._config.set(
                                section=section, option=option, value=value
                            )
                        else:
                            logger.debug(
                                f"New direct assignment {section=} {option=} ignored"
                            )


class NewOptionsFromEnvStrategy(Strategy):
    def execute(self):
        self.override_env(create_new_options=True)
        self.override_direct(create_new_options=False)


class NewOptionsFromDirectStrategy(Strategy):
    def execute(self):
        self.override_env(create_new_options=False)
        self.override_direct(create_new_options=True)


class NewOptionsFromDirectAndEnvStrategy(Strategy):
    def execute(self):
        self.override_env(create_new_options=True)
        self.override_direct(create_new_options=True)


class NoNewOptionsStrategy(Strategy):
    def execute(self):
        self.override_env(create_new_options=False)
        self.override_direct(create_new_options=False)


class OverrideStrategies(enum.Enum):
    NEW_OPTIONS_FROM_ENV = NewOptionsFromEnvStrategy
    NEW_OPTIONS_FROM_DIRECT = NewOptionsFromDirectStrategy
    NEW_OPTIONS_FROM_DIRECT_AND_ENV = NewOptionsFromDirectAndEnvStrategy
    NO_NEW_OPTIONS = NoNewOptionsStrategy



__version__ = "0.5.0"

import logging

from configparser_override.configparser_override import ConfigParserOverride
from configparser_override.convert import ConfigConverter

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = ["__version__", "ConfigParserOverride", "ConfigConverter"]



import configparser
import platform
from dataclasses import dataclass

from configparser_override import ConfigParserOverride, __version__
from tests._constants import TEST_ENV_PREFIX


def test_initialization():
    parser = ConfigParserOverride(env_prefix=TEST_ENV_PREFIX)
    assert parser.env_prefix == TEST_ENV_PREFIX
    assert isinstance(parser.config, configparser.ConfigParser)


def test_read_config_file(config_file):
    parser = ConfigParserOverride()
    config = parser.read(filenames=config_file)

    assert config["SECTION1"]["key1"] == "value1"
    assert config["SECTION1"]["key2"] == "value2"
    assert config["SECTION2"]["key3"] == "value3"


def test_env_override_with_prefix(monkeypatch, config_file):
    monkeypatch.setenv(f"{TEST_ENV_PREFIX}SECTION1__KEY1", "override1")
    monkeypatch.setenv(f"{TEST_ENV_PREFIX}SECTION2__KEY3", "override3")

    parser = ConfigParserOverride(env_prefix=TEST_ENV_PREFIX)
    config = parser.read(filenames=config_file)

    assert config["SECTION1"]["key1"] == "override1"
    assert config["SECTION1"]["key2"] == "value2"  # Not overridden
    assert config["SECTION2"]["key3"] == "override3"


def test_default_section_override_with_env(monkeypatch, config_file_with_default):
    monkeypatch.setenv(f"{TEST_ENV_PREFIX}DEFAULT_KEY", "override_default")

    parser = ConfigParserOverride(env_prefix=TEST_ENV_PREFIX)
    config = parser.read(filenames=config_file_with_default)

    assert config.defaults()["default_key"] == "override_default"


def test_version_exists():
    assert isinstance(__version__, str)


def test_direct_override_takes_precedence_over_env(monkeypatch, config_file):
    monkeypatch.setenv(f"{TEST_ENV_PREFIX}SECTION1__KEY1", "env_override_value1")

    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX, SECTION1__key1="direct_override_value1"
    )
    config = parser.read(filenames=config_file)

    assert config["SECTION1"]["key1"] == "direct_override_value1"
    assert config["SECTION1"]["key2"] == "value2"  # Not overridden
    assert config["SECTION2"]["key3"] == "value3"  # Not overridden


def test_direct_override_from_file(config_file):
    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX, SECTION1__key1="direct_override_value1"
    )
    config = parser.read(filenames=config_file)

    assert config["SECTION1"]["key1"] == "direct_override_value1"
    assert config["SECTION1"]["key2"] == "value2"  # Not overridden
    assert config["SECTION2"]["key3"] == "value3"  # Not overridden


def test_combined_env_and_direct_override(monkeypatch, config_file):
    monkeypatch.setenv(f"{TEST_ENV_PREFIX}SECTION1__KEY2", "env_override_value2")

    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX, SECTION1__key1="direct_override_value1"
    )
    config = parser.read(filenames=config_file)

    assert config["SECTION1"]["key1"] == "direct_override_value1"
    assert config["SECTION1"]["key2"] == "env_override_value2"
    assert config["SECTION2"]["key3"] == "value3"  # Not overridden


def test_direct_override_in_default_section(config_file_with_default):
    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX, default_key="direct_override_default_value"
    )
    config = parser.read(filenames=config_file_with_default)

    assert config.defaults()["default_key"] == "direct_override_default_value"


def test_config_parser_override_with_combined_overrides(monkeypatch):
    config_override = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX,
        SECTION1__option1="override_value1",
    )

    monkeypatch.setenv(f"{TEST_ENV_PREFIX}SECTION2__OPTION2", "env_value2")
    config = config_override.read([])
    assert config.get("section1", "option1") == "override_value1"
    assert config.get("section2", "option2") == "env_value2"


def test_config_parser_override_with_combined_overrides_case_sensitive(monkeypatch):
    config_override = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX,
        case_sensitive_overrides=True,
        SECTION1__option1="override_value1",
    )

    monkeypatch.setenv(f"{TEST_ENV_PREFIX}SECTION2__OPTION2", "env_value2")
    config = config_override.read([])
    assert config.get("SECTION1", "option1") == "override_value1"
    assert config.get("SECTION2", "option2") == "env_value2"


def test_case_insensitive_env_override(monkeypatch, config_file):
    monkeypatch.setenv(
        f"{TEST_ENV_PREFIX}section1__key1", "env_override_value1"
    )  # ENV case name is test

    parser = ConfigParserOverride(env_prefix=TEST_ENV_PREFIX)
    config = parser.read(filenames=config_file)
    assert config["SECTION1"]["key1"] == "env_override_value1"
    assert config["SECTION1"]["key2"] == "value2"  # Not overridden
    assert config["SECTION2"]["key3"] == "value3"  # Not overridden


def test_case_insensitive_only_option_env_override(monkeypatch, config_file):
    monkeypatch.setenv(
        f"{TEST_ENV_PREFIX}SECTION1__key1", "env_override_value1"
    )  # ENV case name is test

    parser = ConfigParserOverride(env_prefix=TEST_ENV_PREFIX)
    config = parser.read(filenames=config_file)
    assert config["SECTION1"]["key1"] == "env_override_value1"
    assert config["SECTION1"]["key2"] == "value2"  # Not overridden
    assert config["SECTION2"]["key3"] == "value3"  # Not overridden


def test_case_insensitive_upper_direct_override(config_file):
    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX, SECTION1__KEY1="direct_override_value1"
    )
    config = parser.read(filenames=config_file)

    assert config["SECTION1"]["key1"] == "direct_override_value1"
    assert config["SECTION1"]["key2"] == "value2"  # Not overridden
    assert config["SECTION2"]["key3"] == "value3"  # Not overridden


def test_case_insensitive_lower_direct_override(config_file):
    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX, section1__KEY1="direct_override_value1"
    )
    config = parser.read(filenames=config_file)

    assert config["SECTION1"]["key1"] == "direct_override_value1"
    assert config["SECTION1"]["key2"] == "value2"  # Not overridden
    assert config["SECTION2"]["key3"] == "value3"  # Not overridden


def test_case_insensitive_lower_direct_multi_override(config_file):
    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX,
        section1__KEY1="direct_override_value1",
        section2__KEY3="direct_override_value3",
    )
    config = parser.read(filenames=config_file)

    assert config["SECTION1"]["key1"] == "direct_override_value1"
    assert config["SECTION1"]["key2"] == "value2"  # Not overridden
    assert config["SECTION2"]["key3"] == "direct_override_value3"


def test_case_insensitive_lower_direct_not_new(config_file):
    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX,
        create_new_from_direct=False,
        section3__KEY3="direct_override_value3",
        section3__KEY4="direct_override_value4",
    )
    config = parser.read(filenames=config_file)

    assert config["SECTION1"]["key1"] == "value1"  # Not overridden
    assert config["SECTION1"]["key2"] == "value2"  # Not overridden
    assert config["SECTION2"]["key3"] == "value3"  # Not overridden
    assert config.has_section("section3") is False


def test_combined_case_insensitive_overrides(monkeypatch, config_file):
    monkeypatch.setenv(f"{TEST_ENV_PREFIX}section1__KEY2", "env_override_value2")

    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX,
        SECTION1__KEY1="direct_override_value1",
        create_new_from_direct=True,
        create_new_from_env_prefix=True,
    )
    config = parser.read(filenames=config_file)

    assert config["SECTION1"]["key1"] == "direct_override_value1"
    assert config["SECTION1"]["key2"] == "env_override_value2"
    assert config["SECTION2"]["key3"] == "value3"  # Not overridden


def test_combined_case_sensitive_overrides(monkeypatch, config_file):
    monkeypatch.setenv(f"{TEST_ENV_PREFIX}section1__KEY2", "env_override_value2")

    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX,
        SECTION1__KEY1="direct_override_value1",
        create_new_from_direct=True,
        create_new_from_env_prefix=True,
        case_sensitive_overrides=True,
    )
    config = parser.read(filenames=config_file)

    p = platform.system()
    if p == "Windows":
        assert config["SECTION1"]["key1"] == "direct_override_value1"
        assert config["SECTION1"]["key2"] == "env_override_value2"
        assert config["SECTION2"]["key3"] == "value3"  # Not overridden
    elif p == "Linux" or p == "Darwin":
        assert config["SECTION1"]["key1"] == "direct_override_value1"
        assert config["SECTION1"]["key2"] == "value2"
        assert config["section1"]["key2"] == "env_override_value2"
        assert config["SECTION2"]["key3"] == "value3"  # Not overridden


def test_combined_case_sensitive_overrides_no_new(monkeypatch, config_file):
    monkeypatch.setenv(f"{TEST_ENV_PREFIX}section1__KEY2", "env_override_value2")

    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX,
        SECTION1__KEY1="direct_override_value1",
        SECTIONNONE__KEY1="direct_override_value_none",
        create_new_from_direct=False,
        create_new_from_env_prefix=False,
        case_sensitive_overrides=True,
    )
    config = parser.read(filenames=config_file)

    p = platform.system()
    if p == "Windows":
        assert config["SECTION1"]["key1"] == "direct_override_value1"
        assert config["SECTION1"]["key2"] == "env_override_value2"
        assert config["SECTION2"]["key3"] == "value3"  # Not overridden
        assert not config.has_section("SECTIONNONE")
        assert not config.has_option("SECTIONNONE", "KEY1")
    elif p == "Linux" or p == "Darwin":
        assert config["SECTION1"]["key1"] == "direct_override_value1"
        assert config["SECTION1"]["key2"] == "value2"
        assert config["SECTION2"]["key3"] == "value3"  # Not overridden
        assert not config.has_section("section1")
        assert not config.has_section("SECTIONNONE")
        assert not config.has_option("SECTIONNONE", "KEY1")


def test_combined_overrides_with_default_section(monkeypatch, config_file_with_default):
    monkeypatch.setenv(f"{TEST_ENV_PREFIX}DEFAULT_KEY", "env_override_default_value")
    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX, default_key="direct_override_default_value"
    )
    config = parser.read(filenames=config_file_with_default)

    assert config.defaults()["default_key"] == "direct_override_default_value"


def test_custom_config_parser():
    custom_parser = configparser.ConfigParser()
    custom_parser.add_section("CUSTOM")
    custom_parser.set("CUSTOM", "key", "custom_value")

    parser = ConfigParserOverride(config_parser=custom_parser)
    config = parser.config

    assert config["CUSTOM"]["key"] == "custom_value"


def test_custom_config_parser_with_combined_overrides(monkeypatch):
    custom_parser = configparser.ConfigParser()
    custom_parser.add_section("CUSTOM")
    custom_parser.set("CUSTOM", "key", "custom_value")

    config_override = ConfigParserOverride(
        config_parser=custom_parser,
        env_prefix=TEST_ENV_PREFIX,
        create_new_from_env_prefix=True,
        create_new_from_direct=True,
        CUSTOM__key="override_value",
    )

    monkeypatch.setenv(f"{TEST_ENV_PREFIX}CUSTOM__NEWKEY", "env_value")
    config = config_override.read([])

    assert config.get("CUSTOM", "key") == "override_value"
    assert config.get("CUSTOM", "newkey") == "env_value"


def test_custom_config_parser_with_file(config_file):
    custom_parser = configparser.ConfigParser()
    custom_parser.add_section("CUSTOM")
    custom_parser.set("CUSTOM", "key", "custom_value")

    parser = ConfigParserOverride(
        config_parser=custom_parser,
        env_prefix=TEST_ENV_PREFIX,
        SECTION1__key1="override_value",
    )
    config = parser.read(filenames=config_file)

    assert config["CUSTOM"]["key"] == "custom_value"
    assert config["SECTION1"]["key1"] == "override_value"
    assert config["SECTION1"]["key2"] == "value2"  # Not overridden
    assert config["SECTION2"]["key3"] == "value3"  # Not overridden


def test_config_parser_with_default_section_new_direct(config_file_with_default):
    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX,
        create_new_from_env_prefix=False,
        new_default_key="direct_override_default_value",
    )
    config = parser.read(filenames=config_file_with_default)

    assert config.defaults()["new_default_key"] == "direct_override_default_value"
    assert config.defaults()["default_key"] == "default_value"
    assert config["DEFAULT"]["default_key"] == "default_value"


def test_config_parser_with_default_section_env_override(
    monkeypatch, config_file_with_default
):
    monkeypatch.setenv(f"{TEST_ENV_PREFIX}DEFAULT_KEY", "env_override_default_value")

    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX,
        create_new_from_env_prefix=False,
    )
    config = parser.read(filenames=config_file_with_default)

    assert config.defaults()["default_key"] == "env_override_default_value"
    assert config["DEFAULT"]["default_key"] == "env_override_default_value"


def test_custom_config_parser_with_default_section(config_file_with_default):
    custom_parser = configparser.ConfigParser()
    custom_parser.set("DEFAULT", "default_key", "custom_default_value")

    parser = ConfigParserOverride(
        config_parser=custom_parser,
        env_prefix=TEST_ENV_PREFIX,
        default_key="direct_override_default_value",
    )
    config = parser.read(filenames=config_file_with_default)

    assert config.defaults()["default_key"] == "direct_override_default_value"
    assert config["DEFAULT"]["default_key"] == "direct_override_default_value"


def test_custom_config_parser_with_custom_default_section(
    config_file_with_custom_default,
):
    custom_parser = configparser.ConfigParser(default_section="COMMON")

    parser = ConfigParserOverride(
        config_parser=custom_parser,
    )
    config = parser.read(filenames=config_file_with_custom_default)

    assert config.defaults()["default_key1"] == "default_value1"
    assert config["COMMON"]["default_key1"] == "default_value1"


def test_custom_config_parser_with_custom_default_section_override(
    monkeypatch, config_file_with_custom_default
):
    custom_parser = configparser.ConfigParser(default_section="COMMON")
    monkeypatch.setenv(f"{TEST_ENV_PREFIX}DEFAULT_KEY1", "env_value1")

    parser = ConfigParserOverride(
        config_parser=custom_parser,
        env_prefix=TEST_ENV_PREFIX,
        default_key2="direct_override_default_value2",
    )
    config = parser.read(filenames=config_file_with_custom_default)

    assert config["COMMON"]["default_key2"] == "direct_override_default_value2"
    assert config["COMMON"]["default_key1"] == "env_value1"


def test_custom_optionxform_insensetive(monkeypatch, config_file):
    monkeypatch.setenv(f"{TEST_ENV_PREFIX}SECTION1__KEY2", "env_override_value2")

    def customform(optionstr: str) -> str:
        return optionstr

    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX,
        optionxform=customform,
        section1__key1="direct_override_value1",
    )
    config = parser.read(filenames=config_file)

    assert config["SECTION1"]["key1"] == "direct_override_value1"
    assert config["SECTION1"]["KEY2"] == "env_override_value2"
    assert config["SECTION2"]["key3"] == "value3"  # Not overridden


def test_custom_optionxform_sensetive(monkeypatch, config_file):
    monkeypatch.setenv(f"{TEST_ENV_PREFIX}SECTION1__KEY2", "env_override_value2")

    def customform(optionstr: str) -> str:
        return optionstr

    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX,
        optionxform=customform,
        SECTION1__KEY1="direct_override_value1",
    )
    config = parser.read(filenames=config_file)

    assert config["SECTION1"]["KEY1"] == "direct_override_value1"
    assert config["SECTION1"]["KEY2"] == "env_override_value2"
    assert config["SECTION2"]["key3"] == "value3"  # Not overridden


def test_combined_case_sensitive_overrides_plus_prefix(monkeypatch, config_file):
    monkeypatch.setenv(
        f"{TEST_ENV_PREFIX.lower()}SECTION1__key2", "env_override_value2"
    )

    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX.lower(),
        SECTION1__KEY1="direct_override_value1",
        create_new_from_direct=False,
        create_new_from_env_prefix=False,
        case_sensitive_overrides=True,
    )
    config = parser.read(filenames=config_file)

    assert config["SECTION1"]["key1"] == "direct_override_value1"
    assert config["SECTION1"]["key2"] == "env_override_value2"
    assert config["SECTION2"]["key3"] == "value3"  # Not overridden


def test_config_to_dataclass(config_file):
    @dataclass
    class Section1:
        key1: str
        key2: str

    @dataclass
    class Section2:
        key3: str

    @dataclass
    class ConfigFile:
        SECTION1: Section1
        SECTION2: Section2

    parser = ConfigParserOverride()
    parser.read(filenames=config_file)

    dataclass_config = parser.to_dataclass(ConfigFile)
    assert dataclass_config.SECTION1.key1 == "value1"
    assert dataclass_config.SECTION1.key2 == "value2"
    assert dataclass_config.SECTION2.key3 == "value3"


def test_direct_override_and_direct_ignore(config_file):
    parser = ConfigParserOverride(
        create_new_from_direct=False,
        section1__key1="direct1",
        sectionMiss="not applied",
    )
    config = parser.read(filenames=config_file)

    assert config["SECTION1"]["key1"] == "direct1"
    assert config["SECTION1"]["key2"] == "value2"
    assert config["SECTION2"]["key3"] == "value3"



import configparser

import pytest

from configparser_override._override_strategy import (
    NewOptionsFromDirectAndEnvStrategy,
    NewOptionsFromDirectStrategy,
    NewOptionsFromEnvStrategy,
    NoNewOptionsStrategy,
)
from configparser_override._strategy_factory import StrategyFactory
from configparser_override.exceptions import OverrideStrategyNotImplementedError
from tests._constants import TEST_ENV_PREFIX


def test_strategy_factory_no_prefix_no_new():
    config = configparser.ConfigParser()
    factory = StrategyFactory(config, "", False, False, {})
    strategy = factory.get_strategy()
    assert isinstance(strategy, NoNewOptionsStrategy)


def test_strategy_factory_no_prefix_new_direct():
    config = configparser.ConfigParser()
    factory = StrategyFactory(config, "", False, True, {})
    strategy = factory.get_strategy()
    assert isinstance(strategy, NewOptionsFromDirectStrategy)


def test_strategy_factory_prefix_no_new():
    config = configparser.ConfigParser()
    factory = StrategyFactory(config, TEST_ENV_PREFIX, False, False, {})
    strategy = factory.get_strategy()
    assert isinstance(strategy, NoNewOptionsStrategy)


def test_strategy_factory_prefix_new_direct():
    config = configparser.ConfigParser()
    factory = StrategyFactory(config, TEST_ENV_PREFIX, False, True, {})
    strategy = factory.get_strategy()
    assert isinstance(strategy, NewOptionsFromDirectStrategy)


def test_strategy_factory_prefix_new_env():
    config = configparser.ConfigParser()
    factory = StrategyFactory(config, TEST_ENV_PREFIX, True, False, {})
    strategy = factory.get_strategy()
    assert isinstance(strategy, NewOptionsFromEnvStrategy)


def test_strategy_factory_prefix_new_env_new_direct():
    config = configparser.ConfigParser()
    factory = StrategyFactory(config, TEST_ENV_PREFIX, True, True, {})
    strategy = factory.get_strategy()
    assert isinstance(strategy, NewOptionsFromDirectAndEnvStrategy)


def test_strategy_factory_raises_not_implemented_error():
    config = configparser.ConfigParser()
    factory = StrategyFactory(config, "", True, "s", {})  # type: ignore
    with pytest.raises(OverrideStrategyNotImplementedError):
        factory.get_strategy()



from dataclasses import dataclass
from typing import Callable, Optional

import pytest

from configparser_override import ConfigParserOverride
from configparser_override.convert import ConfigConverter
from configparser_override.exceptions import ConversionError, LiteralEvalMiscast


@pytest.fixture()
def config_file_simple_types(tmp_path):
    config_content = """
    [DEFAULT]
    allkey = string

    [section1]
    key1 = 123
    key2 = string

    [section2]
    key3 = 1.2
    """
    config_path = tmp_path / "config.ini"
    config_path.write_text(config_content)
    return str(config_path)


@dataclass
class DefaultSection:
    allkey: str


@dataclass
class Section1:
    key1: int
    key2: str
    allkey: str


@dataclass
class Section2:
    key3: float
    allkey: str


@dataclass
class ConfigFileSimpleTypes:
    DEFAULT: DefaultSection
    section1: Section1
    section2: Section2


@pytest.fixture()
def config_file_complex_types(tmp_path):
    config_content = """
    [DEFAULT]
    allkey = byte
    optionalallkey = 123

    [section1]
    key1 = [1,2,3,4,5,6]
    key2 = {"key": "value", "key2": "value2"}

    [section2]
    key3 = [1.2,1.4]
    """
    config_path = tmp_path / "config.ini"
    config_path.write_text(config_content)
    return str(config_path)


@dataclass
class ComplexDefaultSection:
    allkey: bytes
    optionalallkey: Optional[str | int]


@dataclass
class ComplexSection1:
    key1: list[int | str]
    key2: dict[str, str]
    allkey: bytes
    optionalallkey: Optional[str | int]


@dataclass
class ComplexSection2:
    key3: list[float]
    allkey: bytes
    optionalallkey: Optional[str | int]
    key4: Optional[float] = None


@dataclass
class ConfigFileComlexTypes:
    DEFAULT: ComplexDefaultSection
    section1: ComplexSection1
    section2: ComplexSection2


@pytest.fixture()
def config_file_complex_types_nested(tmp_path):
    config_content = """
    [DEFAULT]
    allkey = byte
    optionalallkey = 123

    [section1]
    key1 = [[1,2,3],[4,5,6]]
    key2 = {"key": {"nestedkey": "value"}, "key2": {"nestedkey2": "value2"}}

    [section2]
    key3 = ["true",1,"yes","no",0,"false"]

    [section3]
    key4 = {"string1","string2"}
    key5 = [{"string1"},{"string2"}]
    key6 = ({"123j"},{"4+2j"})
    """
    config_path = tmp_path / "config.ini"
    config_path.write_text(config_content)
    return str(config_path)


@dataclass
class ComplexNestedDefaultSection:
    allkey: bytes
    optionalallkey: Optional[str | int]


@dataclass
class ComplexNestedSection1:
    key1: list[list[int | str]]
    key2: dict[str, dict[str, str]]
    allkey: bytes
    optionalallkey: Optional[str | int]


@dataclass
class ComplexNestedSection2:
    key3: list[Optional[bool | str]]
    allkey: bytes
    optionalallkey: Optional[str | int]
    key4: Optional[float] = None


@dataclass
class ComplexNestedSection3:
    key4: set[str]
    key5: list[set[str]]
    key6: tuple[set[complex]]


@dataclass
class ConfigFileComlexNestedTypes:
    DEFAULT: ComplexNestedDefaultSection
    section1: ComplexNestedSection1
    section2: ComplexNestedSection2
    section3: ComplexNestedSection3


def test_simple_config_to_dataclass(config_file_simple_types):
    parser = ConfigParserOverride()
    parser.read(filenames=config_file_simple_types)

    dataclass_rep = ConfigConverter(parser.config).config_to_dataclass(
        ConfigFileSimpleTypes
    )
    assert dataclass_rep.DEFAULT.allkey == "string"
    assert dataclass_rep.section1.allkey == "string"
    assert dataclass_rep.section1.key1 == 123
    assert dataclass_rep.section1.key2 == "string"
    assert dataclass_rep.section2.key3 == 1.2


def test_complex_config_to_dataclass(config_file_complex_types):
    parser = ConfigParserOverride()
    parser.read(filenames=config_file_complex_types)

    dataclass_rep = ConfigConverter(parser.config).config_to_dataclass(
        ConfigFileComlexTypes
    )
    assert dataclass_rep.DEFAULT.allkey == b"byte"
    assert dataclass_rep.section1.allkey == b"byte"
    assert dataclass_rep.section1.key1 == [1, 2, 3, 4, 5, 6]
    assert dataclass_rep.section1.key2 == {
        "key": "value",
        "key2": "value2",
    }
    assert dataclass_rep.section2.key3 == [1.2, 1.4]


def test_complex_nested_config_to_dataclass(config_file_complex_types_nested):
    parser = ConfigParserOverride()
    parser.read(filenames=config_file_complex_types_nested)

    dataclass_rep = ConfigConverter(parser.config).config_to_dataclass(
        ConfigFileComlexNestedTypes
    )
    assert dataclass_rep.DEFAULT.allkey == b"byte"
    assert dataclass_rep.section1.allkey == b"byte"
    assert dataclass_rep.section1.key1 == [[1, 2, 3], [4, 5, 6]]
    assert dataclass_rep.section1.key2 == {
        "key": {"nestedkey": "value"},
        "key2": {"nestedkey2": "value2"},
    }
    assert dataclass_rep.section2.key3 == [True, True, True, False, False, False]
    assert dataclass_rep.section3.key4 == {"string1", "string2"}
    assert dataclass_rep.section3.key5 == [{"string1"}, {"string2"}]
    assert dataclass_rep.section3.key6 == ({123j}, {4 + 2j})


def test_config_to_dataclass_custom_bools():
    custom_booleans = {"cool": True, "not cool": False}

    @dataclass
    class Sect1:
        true: bool
        false: bool

    @dataclass
    class CustomBools:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__true="cool", sect1__false="not cool")
    parser.read(filenames=[])

    dataclass_rep = ConfigConverter(
        parser.config, boolean_states=custom_booleans
    ).config_to_dataclass(CustomBools)

    assert dataclass_rep.sect1.false is False
    assert dataclass_rep.sect1.true is True


def test_config_to_dataclass_bools_not_valid():
    @dataclass
    class Sect1:
        true: bool
        false: bool

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(
        sect1__true="notValidTrue", sect1__false="notValidFalse"
    )
    parser.read(filenames=[])

    with pytest.raises(ValueError):
        ConfigConverter(parser.config).config_to_dataclass(C)


def test_missing_key_in_config():
    @dataclass
    class Sect1:
        key: str
        key132: str

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key="ok")
    parser.read(filenames=[])

    with pytest.raises(AttributeError):
        ConfigConverter(parser.config).config_to_dataclass(C)


def test_unsupported_type():
    @dataclass
    class Sect1:
        key: Callable

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key="ok")
    parser.read(filenames=[])

    with pytest.raises(ValueError):
        ConfigConverter(parser.config).config_to_dataclass(C)


def test_list_member_conversion_error():
    @dataclass
    class Sect1:
        key: list[int]

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key='["ok"]')
    parser.read(filenames=[])

    with pytest.raises(ConversionError):
        ConfigConverter(parser.config).config_to_dataclass(C)


def test_set_member_conversion_error():
    @dataclass
    class Sect1:
        key: set[int]

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key='{"ok"}')
    parser.read(filenames=[])

    with pytest.raises(ConversionError):
        ConfigConverter(parser.config).config_to_dataclass(C)


def test_dict_member_conversion_error():
    @dataclass
    class Sect1:
        key: dict[str, int]

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key='{"ok":"ok"}')
    parser.read(filenames=[])

    with pytest.raises(ConversionError):
        ConfigConverter(parser.config).config_to_dataclass(C)


def test_tuple_member_conversion_error():
    @dataclass
    class Sect1:
        key: tuple[int]

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key='("ok","ok")')
    parser.read(filenames=[])

    with pytest.raises(ConversionError):
        ConfigConverter(parser.config).config_to_dataclass(C)


def test_union_conversion_error():
    @dataclass
    class Sect1:
        key: list[int | float]

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key='["ok","ok"]')
    parser.read(filenames=[])

    with pytest.raises(ConversionError):
        ConfigConverter(parser.config).config_to_dataclass(C)


def test_list_member_literaleval_error():
    @dataclass
    class Sect1:
        key: list[int]

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key='{"ok"}')
    parser.read(filenames=[])

    with pytest.raises(LiteralEvalMiscast):
        ConfigConverter(parser.config).config_to_dataclass(C)


def test_set_member_literaleval_error():
    @dataclass
    class Sect1:
        key: set[int]

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key='["ok"]')
    parser.read(filenames=[])

    with pytest.raises(LiteralEvalMiscast):
        ConfigConverter(parser.config).config_to_dataclass(C)


def test_dict_member_literaleval_error():
    @dataclass
    class Sect1:
        key: dict[str, int]

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key='["ok"]')
    parser.read(filenames=[])

    with pytest.raises(LiteralEvalMiscast):
        ConfigConverter(parser.config).config_to_dataclass(C)


def test_tuple_member_literaleval_error():
    @dataclass
    class Sect1:
        key: tuple[int]

    @dataclass
    class C:
        sect1: Sect1

    parser = ConfigParserOverride(sect1__key='["ok"]')
    parser.read(filenames=[])

    with pytest.raises(LiteralEvalMiscast):
        ConfigConverter(parser.config).config_to_dataclass(C)



import ast
import configparser
import dataclasses
import logging
from types import UnionType
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from configparser_override.exceptions import ConversionError, LiteralEvalMiscast
from configparser_override.types import _dataclass

logger = logging.getLogger(__name__)


def _is_optional_type(type_hint: Any) -> bool:
    """
    Check if a given type hint is an optional type.

    :param type_hint: The type hint to check.
    :type type_hint: Any
    :return: True if the type hint is optional, False otherwise.
    :rtype: bool
    """
    return get_origin(type_hint) in [Union, UnionType] and type(None) in get_args(
        type_hint
    )


class ConfigConverter:
    """
    A class to convert configuration data from a ConfigParser object to a dictionary
    or dataclass.

    :param config: The configuration parser object.
    :type config: configparser.ConfigParser
    :param boolean_states: Optional mapping of custom boolean states,
        defaults to None and uses the internal mapping of the ConfigParser object.
    :type boolean_states: Optional[Mapping[str, bool]], optional
    """

    def __init__(
        self,
        config: configparser.ConfigParser,
        boolean_states: Optional[Mapping[str, bool]] = None,
    ) -> None:
        self.config = config
        if boolean_states:
            self.boolean_states = boolean_states
        else:
            self.boolean_states = self.config.BOOLEAN_STATES

    def config_to_dict(self) -> dict[str, dict[str, str]]:
        """
        Convert the configuration data to a nested dictionary.

        :return: The configuration data as a dictionary.
        :rtype: dict[str, dict[str, str]]
        """
        config_dict: dict[str, dict[str, str]] = {}
        for sect in self.config.sections():
            # Add nested sections
            config_dict[sect] = {}
            for opt in self.config.options(sect):
                config_dict[sect][opt] = self.config.get(section=sect, option=opt)
        # Add default nested section
        config_dict[self.config.default_section] = {}
        for opt in self.config.defaults():
            config_dict[self.config.default_section][opt] = self.config.get(
                section=self.config.default_section, option=opt
            )
        return config_dict

    def config_to_dataclass(self, dataclass: _dataclass) -> _dataclass:
        """
        Convert the configuration data to a dataclass instance.

        :param dataclass: The dataclass type to convert the configuration data into.
        :type dataclass: _dataclass
        :return: An instance of the dataclass populated with the configuration data.
        :rtype: _dataclass
        """
        config_dict = self.config_to_dict()
        return self._dict_to_dataclass(
            input_dict=config_dict,
            dataclass=dataclass,
        )

    def _dict_to_dataclass(self, input_dict: dict, dataclass: _dataclass) -> _dataclass:
        """
        Convert a dictionary to a dataclass instance.

        :param input_dict: The input dictionary to convert.
        :type input_dict: dict
        :param dataclass: The dataclass type to convert the dictionary into.
        :type dataclass: _dataclass
        :return: An instance of the dataclass populated with the dictionary data.
        :rtype: _dataclass
        :raises ValueError: If the input object is not a dataclass or required
            fields are missing.
        """
        type_hints = get_type_hints(dataclass)

        _dict_with_types: dict[str, Any] = {}
        if not dataclasses.is_dataclass(dataclass):
            raise ValueError(f"object {dataclass} is not a Dataclass")
        for field in dataclasses.fields(dataclass):
            field_name = field.name
            field_type = type_hints[field_name]
            if field_name in input_dict:
                _dict_with_types[field_name] = self._cast_value(
                    value=input_dict[field_name],
                    type_hint=field_type,
                )
            elif not _is_optional_type(field_type):
                raise AttributeError(f"Missing field in source config: {field_name}")
        return dataclass(**_dict_with_types)

    def _cast_value(self, value: Any, type_hint: Any) -> Any:
        """
        Trey to cast a value to a given type hint.

        :param value: The value to cast.
        :type value: Any
        :param type_hint: The type hint to cast the value to.
        :type type_hint: Any
        :return: The value cast to the specified type hint.
        :rtype: Any
        :raises ValueError: If the type hint is unsupported.
        """
        if dataclasses.is_dataclass(type_hint):
            return self._dict_to_dataclass(value, type_hint)  # type: ignore[type-var]
        if type_hint is Any:
            return value
        if type_hint in [int, float, complex, str]:
            return type_hint(value)
        if type_hint is bytes:
            return str(value).encode()
        if type_hint is bool:
            return self._cast_bool(value)
        _origin = get_origin(type_hint)
        if _origin in [list, List]:
            return self._cast_list(value, type_hint)
        if _origin in [dict, Dict]:
            return self._cast_dict(value, type_hint)
        if _origin in [set, Set]:
            return self._cast_set(value, type_hint)
        if _origin in [tuple, Tuple]:
            return self._cast_tuple(value, type_hint)
        if _origin in (Optional, Union, UnionType):
            return self._cast_union(value, type_hint)
        if type_hint is type(None):
            return None
        raise ValueError(f"Unsupported type: {type_hint}")

    def _cast_bool(self, value: Any) -> bool:
        """
        Cast a value to a boolean.

        :param value: The value to cast.
        :type value: Any
        :return: The value cast to a boolean.
        :rtype: bool
        :raises ValueError: If the value cannot be cast to a boolean.
        """
        if str(value).lower() in self.boolean_states:
            return self.boolean_states[str(value).lower()]
        else:
            raise ValueError(f"{value=} not in possible {self.boolean_states=}")

    def _cast_list(self, value: Any, type_hint: Any) -> list:
        """
        Cast a value to a list of a specified type.

        :param value: The value to cast.
        :type value: Any
        :param type_hint: The type hint for the list elements.
        :type type_hint: Any
        :return: The value cast to a list of the specified type.
        :rtype: list
        :raises ConversionError: If the value cannot be cast to a list of hinted types.
        :raises LiteralEvalMiscast: If the value cannot be evaluated to the
            expected type.
        """
        _evaluated_option = ast.literal_eval(value) if isinstance(value, str) else value
        if isinstance(_evaluated_option, list):
            _types = get_args(type_hint)
            for typ in _types:
                try:
                    return [self._cast_value(item, typ) for item in _evaluated_option]
                except Exception as e:
                    logger.debug(f"Faild to cast {value=} into {typ=}, error: {e}")
                    continue
            raise ConversionError(
                f"Not possible to cast {value} into a list of {_types}"
            )
        raise LiteralEvalMiscast(
            f"{value} casted as {type(_evaluated_option)} expected {type_hint}"
        )

    def _cast_set(self, value: Any, type_hint: Any) -> set:
        """
        Cast a value to a set of a specified type.

        :param value: The value to cast.
        :type value: Any
        :param type_hint: The type hint for the set elements.
        :type type_hint: Any
        :return: The value cast to a set of the specified type.
        :rtype: set
        :raises ConversionError: If the value cannot be cast to a set of hinted types.
        :raises LiteralEvalMiscast: If the value cannot be evaluated to the
            expected type.
        """
        _evaluated_option = ast.literal_eval(value) if isinstance(value, str) else value
        if isinstance(_evaluated_option, set):
            _types = get_args(type_hint)
            for typ in _types:
                try:
                    return {self._cast_value(item, typ) for item in _evaluated_option}
                except Exception as e:
                    logger.debug(f"Faild to cast {value=} into {typ=}, error: {e}")
                    continue
            raise ConversionError(
                f"Not possible to cast {value} into a set of {_types}"
            )
        raise LiteralEvalMiscast(
            f"{value} casted as {type(_evaluated_option)} expected {type_hint}"
        )

    def _cast_tuple(self, value: Any, type_hint: Any) -> tuple:
        """
        Cast a value to a tuple of a specified type.

        :param value: The value to cast.
        :type value: Any
        :param type_hint: The type hint for the tuple elements.
        :type type_hint: Any
        :return: The value cast to a tuple of the specified type.
        :rtype: tuple
        :raises ConversionError: If the value cannot be cast to a tuple of hinted types.
        :raises LiteralEvalMiscast: If the value cannot be evaluated to the
            expected type.
        """
        _evaluated_option = ast.literal_eval(value) if isinstance(value, str) else value
        if isinstance(_evaluated_option, tuple):
            _types = get_args(type_hint)
            for typ in _types:
                try:
                    return tuple(
                        self._cast_value(item, typ) for item in _evaluated_option
                    )
                except Exception as e:
                    logger.debug(f"Faild to cast {value=} into {typ=}, error: {e}")
                    continue
            raise ConversionError(
                f"Not possible to cast {value} into a tuple of {_types}"
            )
        raise LiteralEvalMiscast(
            f"{value} casted as {type(_evaluated_option)} expected {type_hint}"
        )

    def _cast_dict(self, value: Any, type_hint: Any) -> dict:
        """
        Cast a value to a dictionary of specified types for keys and values.

        :param value: The value to cast.
        :type value: Any
        :param type_hint: The type hint for the dictionary keys and values.
        :type type_hint: Any
        :return: The value cast to a dictionary of the specified types.
        :rtype: dict
        :raises ConversionError: If the value cannot be cast to a dictionary of
            hinted types
        :raises LiteralEvalMiscast: If the value cannot be evaluated to the
            expected type.
        """
        _evaluated_option = ast.literal_eval(value) if isinstance(value, str) else value
        if isinstance(_evaluated_option, dict):
            k_typ, v_typ = get_args(type_hint)
            try:
                return {
                    self._cast_value(k, k_typ): self._cast_value(v, v_typ)
                    for k, v in _evaluated_option.items()
                }
            except Exception as e:
                logger.debug(
                    f"Faild to cast {value=} into {k_typ=}, {v_typ=}, error: {e}"
                )
                raise ConversionError(
                    f"Not possible to cast {value} into a dict of keys of type {k_typ}, and values of type {v_typ}, Error: {e}"
                ) from e
        raise LiteralEvalMiscast(
            f"{value} casted as {type(_evaluated_option)} expected {type_hint}"
        )

    def _cast_union(self, value: Any, type_hint: Any) -> Any:
        """
        Cast a value to one of the types in a union type hint.

        :param value: The value to cast.
        :type value: Any
        :param type_hint: The union type hint.
        :type type_hint: Any
        :return: The value cast to one of the types in the union.
        :rtype: Any
        :raises ConversionError: If the value cannot be cast to any of the types in
            the union.
        """
        for typ in get_args(type_hint):
            try:
                return self._cast_value(value, typ)
            except Exception as e:
                logger.debug(f"Faild to cast {value=} into {typ=}, error: {e}")
                continue
        raise ConversionError(f"Not possible to cast {value} into type {type_hint}")
